#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``beam_transforms`` module contains common functions used by the broker in
Apache Beam jobs.
"""

from apache_beam import DoFn


class ExtractAlertDict(DoFn):
    """Extract the alert packet from a Pub/Sub message (bytes object) and
    return it as a dictionary.
    """
    def process(self, msg):
        from io import BytesIO
        from fastavro import reader

        with BytesIO(msg) as fin:
            alertDicts = [r for r in reader(fin)]
            # list of dicts, generally expected to have length 1
        return alertDicts


class StripCutouts(DoFn):
    """Drop the cutouts from the alert dictionary.
    """
    def __init__(self, schema_map):
        super().__init__()
        self.schema_map = schema_map

    def process(self, alertDict):
        import copy

        schema_map = self.schema_map
        cutouts = [
            schema_map['cutoutScience'],
            schema_map['cutoutTemplate'],
            schema_map['cutoutDifference']
        ]

        if schema_map['SURVEY'] == 'decat':
            alertStripped = copy.deepcopy(alertDict)
            for co in cutouts:
                alertStripped[schema_map['source']].pop(co, None)
                for psource in alertStripped[schema_map['prvSources']]:
                    psource.pop(co, None)

        elif schema_map['SURVEY'] == 'ztf':
            alertStripped = {k:v for k, v in alertDict.items() if k not in cutouts}

        return [alertStripped]


class ExtractDIASource(DoFn):
    """Extract the DIA source` fields and information needed for provenance
    from the alertDict.
    """
    def __init__(self, schema_map):
        super().__init__()
        self.schema_map = schema_map

    def process(self, alertDict):
        SURVEY = self.schema_map['SURVEY']
        if SURVEY == 'ztf':
            source = self.process_ztf(alertDict)
        elif SURVEY == 'decat':
            source = self.process_decat(alertDict)

        return source

    def process_decat(self, alertDict):
        # get source
        dup_cols = ['ra','dec']  # names duplicated in object and source levels
        sourcename = lambda x: x if x not in dup_cols else f'source_{x}'
        src = {sourcename(k):v for k,v in alertDict['triggersource'].items()}

        # get info for provenance
        notmetakeys = ['triggersource', 'sources']
        metadict = {k:v for k,v in alertDict.items() if k not in notmetakeys}

        # get string of previous sources' sourceid, comma-separated
        if alertDict['sources'] is not None:
            tmp = [ps['sourceid'] for ps in alertDict['sources']]
            prv_sources = ','.join([f'{sid}' for sid in tmp if sid is not None])
        else:
            prv_sources = None

        # package it up and return
        source = {**metadict, **src, 'sources_sourceids': prv_sources}
        return [source]

    def process_ztf(self, alertDict):
        # get candidate
        dup_cols = ['candid']  # candid is repeated, drop the one nested here
        cand = {k:v for k,v in alertDict['candidate'].items() if k not in dup_cols}

        # get info for provenance
        metakeys = ['schemavsn', 'publisher', 'objectId', 'candid']
        metadict = {k:v for k,v in alertDict.items() if k in metakeys}

        # get string of previous candidates' candid, comma-separated
        if alertDict['prv_candidates'] is not None:
            tmp = [pc['candid'] for pc in alertDict['prv_candidates']]
            prv_candids = ','.join([f'{cid}' for cid in tmp if cid is not None])
        else:
            prv_candids = None

        # package it up and return
        candidate = {**metadict, **cand, 'prv_candidates_candids': prv_candids}
        return [candidate]

class FormatDictForPubSub(DoFn):
    def process(self, alertDict):
        """Converts alert packet dictionaries to format suitable for WriteToPubSub().

        Returns a `PubsubMessage` object with custom attributes attached.
        """
        from apache_beam.io.gcp.pubsub import PubsubMessage
        import json

        # collect message attributes
        if 'alert' in alertDict.keys():
            a = alertDict['alert']
        else:
            a = alertDict
        attrs = {'objectId': str(a['objectId']), 'candid': str(a['candid'])}

        # convert dict -> bytes
        alert_bytes = json.dumps(alertDict).encode('utf-8')

        # package the message
        msg = PubsubMessage(alert_bytes, attrs)

        return [msg]



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""Tools to used by the Pitt-Google Broker software.
"""

# import pkg_resources  # part of setuptools

from . import beam_transforms, consumer_sim, data_utils, gcp_utils, schema_maps


# __version__ = pkg_resources.require("pgb_broker_utils")[0].version



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``schema_maps`` module loads and returns schema maps stored in
Cloud Storage.
"""

from google.cloud import storage
# from typing import List, Tuple, Optional, Union, Generator
from typing import Optional, Union
import yaml

pgb_project_id = 'ardent-cycling-243415'
storage_client = storage.Client()


def load_schema_map(
    survey: str, testid: Union[str, bool], schema: Optional[str] = None
) -> dict:
    """
    Args:
        survey: Name of the survey associated with the broker instance.
                Along with the `testid`, this determines which GCS bucket the
                schema map will be loaded from. If `schema` is not provided,
                this will also determine which schema map is returned.
        testid: Name of the testid associated with the broker instance.
                Along with the `survey`, this determines which GCS bucket the
                schema map will be loaded from.
                If this is a bool, it must be False (indicates a production instance).
        schema: Survey name of the schema to be returned. If not provided, the
                map corresponding to `survey` will be returned.
    """
    if schema is None: schema = survey

    # load the map from the yaml in cloud storage
    broker_bucket_name = _broker_bucket_name(survey, testid)
    schema_file_name = _schema_file_name(schema)
    blob = storage_client.bucket(broker_bucket_name).get_blob(schema_file_name)
    with blob.open("rt") as f:
        schema_map = yaml.safe_load(f)  # dict

    return schema_map

def _broker_bucket_name(survey, testid):
    if testid in ["False", False]:
        return f'{pgb_project_id}-{survey}-broker_files'
    else:
        return f'{pgb_project_id}-{survey}-broker_files-{testid}'

def _schema_file_name(survey):
    return f'schema_maps/{survey}.yaml'



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""A module for Pitt-Google Broker utilities.
Example setup.py: https://github.com/pypa/sampleproject/blob/main/setup.py

This file is part of the pgb_broker_utils software package.

The pgb_broker_utils package is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The pgb_broker_utils package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License
along with pgb_broker_utils.  If not, see <http://www.gnu.org/licenses/>."""

from setuptools import setup, find_packages
import pathlib


here = pathlib.Path(__file__).parent.resolve()
# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='pgb_broker_utils',
    version='0.2.23',
    description='Tools used by the Pitt-Google astronomical alert broker.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='astronomy, alert broker',

    url='https://github.com/mwvgroup/Pitt-Google-Broker',
    author='Troy Raen',
    author_email='troy.raen@pitt.edu',

    packages=find_packages(),  # Required

    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements

    # python_requires='>=3.6, <4',

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    # package_data={  # Optional
    #     'sample': ['package_data.dat'],
    # },
)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``data_utils`` module contains common functions used to manipulate
survey and broker data.
"""

from astropy.time import Time
import fastavro
from io import BytesIO
import json
import numpy as np
import pandas as pd
from typing import Optional, Tuple, Union


def decode_alert(
    alert_avro: Union[str, bytes],
    return_as: str = 'dict',
    schema_map: Optional[dict] = None,
) -> Union[dict, pd.DataFrame]:
    """Load an alert Avro and return in requested format.

    Wraps `alert_avro_to_dict()` and `alert_dict_to_dataframe()`.

    Args:
        alert_avro:   Either the path of Avro file to load, or
            the bytes encoding the Avro-formated alert.
        return_as: Format the alert will be returned in.
        schema_map: Mapping between survey schema and broker's generic schema.
            Required if `return_as == 'df'`.

    Returns:
        alert packet in requested format
    """
    alert_dict = alert_avro_to_dict(alert_avro)

    if return_as == "dict":
        return alert_dict
    elif return_as == "df":
        return alert_dict_to_dataframe(alert_dict, schema_map)
    else:
        raise ValueError("`return_as` must be one of 'dict' or 'df'.")


def alert_avro_to_dict(alert_avro: Union[str, bytes]) -> dict:
    """Load an alert Avro to a dictionary.

    Args:
        alert_avro:   Either the path of Avro file to load, or
            the bytes encoding the Avro-formated alert.

    Returns:
        alert as a dict
    """
    if type(alert_avro) == str:
        with open(alert_avro, 'rb') as fin:
            alert_list = [r for r in fastavro.reader(fin)]  # list of dicts
    elif type(alert_avro) == bytes:
        try:
            with BytesIO(alert_avro) as fin:
                alert_list = [r for r in fastavro.reader(fin)]  # list of dicts
        except ValueError:
            try:
                # this function is mis-named because here we accept json encoding.
                # consider re-encoding alert packets using original Avro format
                # throughout broker to give users a consistent end-product.
                # then this except block will be unnecessary.
                alert_list = [json.loads(alert_avro.decode("UTF-8"))]  # list of dicts
            except ValueError:
                msg = (
                    "alert_avro is a bytes object, but does not seem to be either "
                    "json or Avro encoded. Cannot decode."
                )
                raise ValueError(msg)
    else:
        msg = (
            "Unable to open alert_avro. "
            "It must be either the path to a valid Avro file as a string, "
            "or bytes encoding an Avro-formated alert."
        )
        raise ValueError(msg)

    # we expect the list to contain exactly 1 entry
    if len(alert_list) == 0:
        raise ValueError('The alert Avro contains 0 valid entries.')
    if len(alert_list) > 1:
        raise ValueError('The alert Avro contains >1 entry.')

    alert_dict = alert_list[0]
    return alert_dict

def alert_dict_to_dataframe(alert_dict: dict, schema_map: dict) -> pd.DataFrame:
    """ Packages an alert into a dataframe.
    Adapted from: https://github.com/ZwickyTransientFacility/ztf-avro-alert/blob/master/notebooks/Filtering_alerts.ipynb
    """
    if type(schema_map) != dict:
        raise ValueError("`schema_map` is not a valid dictionary.")

    dfc = pd.DataFrame(alert_dict[schema_map['source']], index=[0])
    df_prv = pd.DataFrame(alert_dict[schema_map['prvSources']])
    dflc = pd.concat([dfc,df_prv], ignore_index=True)

    # attach some metadata. note this may not be preserved after all operations
    # https://stackoverflow.com/questions/14688306/adding-meta-information-metadata-to-pandas-dataframe
    dflc.objectId = alert_dict[schema_map['objectId']]
    # make sure we haven't onboarded a new survey that has a field named "sourceId"
    assert 'sourceId' not in dflc.keys()
    dflc.sourceId = alert_dict[schema_map['sourceId']]

    return dflc

def mag_to_flux(mag: float, zeropoint: float, magerr: float) -> Tuple[float,float]:
    """ Converts an AB magnitude and its error to fluxes.
    """
    flux = 10 ** ((zeropoint - mag) / 2.5)
    fluxerr = flux * magerr * np.log(10 / 2.5)
    return flux, fluxerr

def jd_to_mjd(jd: float) -> float:
    """ Converts Julian Date to modified Julian Date.
    """
    return Time(jd, format='jd').mjd
